@@ -15,7 +15,11 @@ if (&foreign_check("lvm")) {
 	}
 if (&foreign_check("iscsi-server")) {
 	&foreign_require("iscsi-server");
-	$iscsi_module++;
+	$iscsi_server_module++;
+	}
+if (&foreign_check("iscsi-target")) {
+	&foreign_require("iscsi-target");
+	$iscsi_target_module++;
 	}
 &foreign_require("proc", "proc-lib.pl");
 %access = &get_module_acl();
@@ -47,7 +51,7 @@ if (scalar(@list_disks_partitions_cache)) {
 local (@pscsi, @dscsi, $dscsi_mode);
 if (-r "/proc/scsi/sg/devices" && -r "/proc/scsi/sg/device_strs") {
 	# Get device info from various /proc/scsi files
-	open(DEVICES, "/proc/scsi/sg/devices");
+	open(DEVICES, "</proc/scsi/sg/devices");
 	while(<DEVICES>) {
 		s/\r|\n//g;
 		local @l = split(/\t+/, $_);
@@ -59,7 +63,7 @@ if (-r "/proc/scsi/sg/devices" && -r "/proc/scsi/sg/device_strs") {
 		}
 	close(DEVICES);
 	local $i = 0;
-	open(DEVNAMES, "/proc/scsi/sg/device_strs");
+	open(DEVNAMES, "</proc/scsi/sg/device_strs");
 	while(<DEVNAMES>) {
 		s/\r|\n//g;
 		local @l = split(/\t+/, $_);
@@ -73,7 +77,7 @@ if (-r "/proc/scsi/sg/devices" && -r "/proc/scsi/sg/device_strs") {
 	}
 else {
 	# Check /proc/scsi/scsi for SCSI disk models
-	open(SCSI, "/proc/scsi/scsi");
+	open(SCSI, "</proc/scsi/scsi");
 	local @lines = <SCSI>;
 	close(SCSI);
 	if ($lines[0] =~ /^Attached\s+domains/i) {
@@ -118,7 +122,7 @@ else {
 	}
 
 local (@disks, @devs, $d);
-if (open(PARTS, "/proc/partitions")) {
+if (open(PARTS, "</proc/partitions")) {
 	# The list of all disks can come from the kernel
 	local $sc = 0;
 	while(<PARTS>) {
@@ -150,7 +154,7 @@ if (open(PARTS, "/proc/partitions")) {
 		elsif (/\d+\s+\d+\s+\d+\s+hd([a-z]+)\s/) {
 			# IDE disk (but skip CDs)
 			local $n = $1;
-			if (open(MEDIA, "/proc/ide/hd$n/media")) {
+			if (open(MEDIA, "</proc/ide/hd$n/media")) {
 				local $media = <MEDIA>;
 				close(MEDIA);
 				if ($media =~ /^disk/ && !$_[0]) {
@@ -194,6 +198,14 @@ if (open(PARTS, "/proc/partitions")) {
 			# PV disk from Xen
 			push(@devs, "/dev/$1");
 			}
+		elsif (/\d+\s+\d+\s+\d+\s+(mmcblk\d+)\s/) {
+			# SD card / MMC, seen on Raspberry Pi
+			push(@devs, "/dev/$1");
+			}
+		elsif (/\d+\s+\d+\s+\d+\s+(nvme\d+n\d+)\s/) {
+			# NVME SSD
+			push(@devs, "/dev/$1");
+			}
 		}
 	close(PARTS);
 
@@ -214,6 +226,7 @@ if (!-d "/proc/ide") {
 
 # Get Linux disk ID mapping
 local %id_map;
+local %all_id_map;
 local $id_dir = "/dev/disk/by-id";
 opendir(IDS, $id_dir);
 foreach my $id (readdir(IDS)) {
@@ -221,34 +234,39 @@ foreach my $id (readdir(IDS)) {
 	if ($id_link) {
 		local $id_real = &simplify_path(&resolve_links("$id_dir/$id"));
 		$id_map{$id_real} = $id;
+		$all_id_map{$id_real} ||= [];
+		push(@{$all_id_map{$id_real}}, $id);
 		}
 	}
 closedir(IDS);
 
 # Call fdisk to get partition and geometry information
-local $devs = join(" ", @devs);
+local $qdevs = join(" ", map { quotemeta($_) } @devs);
 local ($disk, $m2);
 if ($has_parted) {
 	open(FDISK, join(" ; ",
-		map { "parted $_ unit cyl print 2>/dev/null || ".
-		      "fdisk -l $_ 2>/dev/null" } @devs)." |");
+		map { "parted ".quotemeta($_)." unit cyl print 2>/dev/null" }
+		    @devs)." |");
 	}
 else {
-	open(FDISK, "fdisk -l $devs 2>/dev/null |");
+	open(FDISK, "fdisk -l -u=cylinders $qdevs 2>/dev/null || fdisk -l $qdevs 2>/dev/null |");
 	}
 while(<FDISK>) {
-	if (/Disk\s+([^ :]+):\s+(\d+)\s+\S+\s+(\d+)\s+\S+\s+(\d+)/ ||
+	if (($m4 = ($_ =~ /Disk\s+([^ :]+):\s+([\d\.]+)\s+(\S+),\s+(\d+)\s+bytes,\s+(\d+)\s+sectors/)) ||
+	    ($m1 = ($_ =~ /Disk\s+([^ :]+):\s+(\d+)\s+\S+\s+(\d+)\s+\S+\s+(\d+)/)) ||
 	    ($m2 = ($_ =~ /Disk\s+([^ :]+):\s+(.*)\s+bytes/)) ||
 	    ($m3 = ($_ =~ /Disk\s+([^ :]+):\s+([0-9\.]+)cyl/))) {
 		# New disk section
 		if ($m3) {
 			# Parted format
+			# Disk /dev/sda: 121601cyl
 			$disk = { 'device' => $1,
 				  'prefix' => $1,
 				  'cylinders' => $2 };
 			}
 		elsif ($m2) {
 			# New style fdisk
+			# Disk /dev/sda: 85.8 GB, 85899345920 bytes
 			$disk = { 'device' => $1,
 				  'prefix' => $1,
 				  'table' => 'msdos', };
@@ -257,6 +275,14 @@ while(<FDISK>) {
 			$disk->{'sectors'} = $2;
 			$disk->{'cylinders'} = $3;
 			}
+		elsif ($m4) {
+			# Even newer disk (sectors/etc come later)
+			# Disk /dev/sda: 10 GiB, 10737418240 bytes, 20971520 sectors
+			$disk = { 'device' => $1,
+				  'prefix' => $1,
+				  'size' => $4,
+				  'table' => 'msdos', };
+			}
 		else {
 			# Old style fdisk
 			$disk = { 'device' => $1,
@@ -290,11 +316,16 @@ while(<FDISK>) {
 			$disk->{'desc'} = &text('select_device', 'Xen', uc($1));
 			$disk->{'type'} = 'ide';
 			}
+		elsif ($disk->{'device'} =~ /\/mmcblk([0-9]+)$/) {
+			# SD-card / MMC
+			$disk->{'desc'} = &text('select_device', 'SD-Card', $1);
+			$disk->{'type'} = 'ide';
+			}
 		elsif ($disk->{'device'} =~ /\/vd([a-z]+)$/) {
 			# KVM virtual disk
 			$disk->{'desc'} = &text('select_device',
 						'VirtIO', uc($1));
-			$disk->{'type'} = 'ide';
+			$disk->{'type'} = 'virtio';
 			}
 		elsif ($disk->{'device'} =~ /\/(scsi\/host(\d+)\/bus(\d+)\/target(\d+)\/lun(\d+)\/disc)/) {
 			# New complete SCSI disk specification
@@ -341,7 +372,7 @@ while(<FDISK>) {
 			# Mylex raid device
 			local ($mc, $md) = ($2, $3);
 			$disk->{'desc'} = &text('select_mylex', $mc, $md);
-			open(RD, "/proc/rd/c$mc/current_status");
+			open(RD, "</proc/rd/c$mc/current_status");
 			while(<RD>) {
 				if (/^Configuring\s+(.*)/i) {
 					$disk->{'model'} = $1;
@@ -359,7 +390,7 @@ while(<FDISK>) {
 			# Compaq RAID device
 			local ($ic, $id) = ($2, $3);
 			$disk->{'desc'} = &text('select_cpq', $ic, $id);
-			open(IDA, -d "/proc/driver/array" ? "/proc/driver/array/ida$ic" : "/proc/driver/cpqarray/ida$ic");
+			open(IDA, -d "/proc/driver/array" ? "</proc/driver/array/ida$ic" : "</proc/driver/cpqarray/ida$ic");
 			while(<IDA>) {
 				if (/^(\S+):\s+(.*)/ && $1 eq "ida$ic") {
 					$disk->{'model'} = $2;
@@ -373,7 +404,7 @@ while(<FDISK>) {
 			# Compaq Smart Array RAID
 			local ($ic, $id) = ($2, $3);
 			$disk->{'desc'} = &text('select_smart', $ic, $id);
-			open(CCI, "/proc/driver/cciss/cciss$ic");
+			open(CCI, "</proc/driver/cciss/cciss$ic");
 			while(<CCI>) {
 				if (/^\s*(\S+):\s*(.*)/ && $1 eq "cciss$ic") {
 					$disk->{'model'} = $2;
@@ -390,6 +421,12 @@ while(<FDISK>) {
 			$disk->{'type'} = 'raid';
 			$disk->{'prefix'} =~ s/disc$/part/g;
 			}
+		elsif ($disk->{'device'} =~ /\/nvme(\d+)n(\d+)$/) {
+			# NVME SSD controller
+			$disk->{'desc'} = &text('select_nvme', "$1", "$2");
+			$disk->{'type'} = 'scsi';
+			$disk->{'prefix'} = $disk->{'device'}.'p';
+			}
 
 		# Work out short name, like sda
 		local $short;
@@ -406,10 +443,19 @@ while(<FDISK>) {
 
 		$disk->{'id'} = $id_map{$disk->{'device'}} ||
 				$id_map{"/dev/$short"};
+		$disk->{'ids'} = $all_id_map{$disk->{'device'}} ||
+				 $all_id_map{"/dev/$short"};
 
 		push(@disks, $disk);
 		}
-	elsif (/^Units\s+=\s+cylinders\s+of\s+(\d+)\s+\*\s+(\d+)/) {
+	elsif (/Geometry:\s+(\d+)\s+heads,\s+(\d+)\s+sectors\/track,\s+(\d+)\s+cylinders/) {
+		# Separate geometry line
+		# Geometry: 255 heads, 63 sectors/track, 1305 cylinders
+		$disk->{'heads'} = $1;
+		$disk->{'sectors'} = $2;
+		$disk->{'cylinders'} = $3;
+		}
+	elsif (/^Units\s*[=:]\s+cylinders\s+of\s+(\d+)\s+\*\s+(\d+)/) {
 		# Unit size for disk from fdisk
 		$disk->{'bytes'} = $2;
 		$disk->{'cylsize'} = $disk->{'heads'} * $disk->{'sectors'} *
@@ -427,8 +473,25 @@ while(<FDISK>) {
 						        $disk->{'sectors'};
 		$disk->{'size'} = $disk->{'cylinders'} * $disk->{'cylsize'};
 		}
+	elsif (/(\/dev\/\S+?(\d+))[ \t*]+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S{1,2})\s+(.*)/) {
+		# Partition within the current disk from fdisk (msdos format)
+		# /dev/sda1 * 1 1306 1306 10G 83 Linux
+		local $part = { 'number' => $2,
+				'device' => $1,
+				'type' => $7,
+				'start' => $3,
+				'end' => $4,
+				'extended' => $7 eq '5' || $6 eq 'f' ? 1 : 0,
+				'index' => scalar(@{$disk->{'parts'}}),
+			 	'edittype' => 1, };
+		$part->{'desc'} = &partition_description($part->{'device'});
+		$part->{'size'} = ($part->{'end'} - $part->{'start'} + 1) *
+				  $disk->{'cylsize'};
+		push(@{$disk->{'parts'}}, $part);
+		}
 	elsif (/(\/dev\/\S+?(\d+))[ \t*]+\d+\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S{1,2})\s+(.*)/ || /(\/dev\/\S+?(\d+))[ \t*]+(\d+)\s+(\d+)\s+(\S+)\s+(\S{1,2})\s+(.*)/) {
-		# Partition within the current disk from fdisk
+		# Partition within the current disk from fdisk (msdos format)
+		# /dev/sda1 * 1 9327 74919096 83 Linux
 		local $part = { 'number' => $2,
 				'device' => $1,
 				'type' => $6,
@@ -443,10 +506,28 @@ while(<FDISK>) {
 				  $disk->{'cylsize'};
 		push(@{$disk->{'parts'}}, $part);
 		}
+	elsif (/(?<dev>\/dev\/\S+?(?<num>\d+))\s+(?<start>\d+)\s+(?<end>\d+)\s+(?<blocks>\d+)\s+(?<size>[0-9\.]+[kMGTP])\s+(?<type>\S.*)/ ||
+	       /(?<dev>\/dev\/\S+?(?<num>\d+))\s+(?<start>\d+)\s+(?<end>\d+)\s+(?<size>[0-9\.]+[kMGTP])\s+(?<type>\S.*)/) {
+		# Partition within the current disk from fdisk (gpt format)
+		local $part = {
+				'number' => "$+{num}",
+				'device' => "$+{dev}",
+				'type' => "$+{type}",
+				'start' => "$+{start}",
+				'end' => "$+{end}",
+				'blocks' => "$+{blocks}",
+				'index' => scalar(@{$disk->{'parts'}}),
+				'dtable' => $disk->{'table'},
+			 	'edittype' => 2, };
+		$part->{'desc'} = &partition_description($part->{'device'});
+		$part->{'size'} = ($part->{'end'} - $part->{'start'} + 1) *
+				  $disk->{'cylsize'};
+		push(@{$disk->{'parts'}}, $part);
+		}
 	elsif (/^\s*(\d+)\s+(\d+)cyl\s+(\d+)cyl\s+(\d+)cyl\s+(primary|logical|extended)\s*(\S*)\s*(\S*)/) {
 		# Partition within the current disk from parted (msdos format)
 		local $part = { 'number' => $1,
-				'device' => $disk->{'device'}.$1,
+				'device' => $disk->{'prefix'}.$1,
 				'type' => $6 || 'ext2',
 				'start' => $2+1,
 				'end' => $3+1,
@@ -463,21 +544,45 @@ while(<FDISK>) {
 				  $disk->{'cylsize'};
 		push(@{$disk->{'parts'}}, $part);
 		}
-	elsif (/^\s*(\d+)\s+(\d+)cyl\s+(\d+)cyl\s+(\d+)cyl\s+(\S+)\s+(\S+)\s+(\S+)/) {
+	elsif (/^\s*(\d+)\s+(\d+)cyl\s+(\d+)cyl\s+(\d+)cyl\s(.*)/) {
 		# Partition within the current disk from parted (gpt format)
 		local $part = { 'number' => $1,
-				'device' => $disk->{'device'}.$1,
-				'type' => $5 || 'ext2',
+				'device' => $disk->{'prefix'}.$1,
 				'start' => $2+1,
 				'end' => $3+1,
 				'blocks' => $4 * $disk->{'cylsize'},
 				'extended' => 0,
-				'raid' => $7 eq 'raid' ? 1 : 0,
 				'index' => scalar(@{$disk->{'parts'}}),
-			        'name' => $6,
 				'edittype' => 0, };
-		$part->{'type'} = 'ext2' if ($part->{'type'} =~ /^ext/);
-		$part->{'type'} = 'raid' if ($part->{'type'} eq 'ext2' &&
+
+		# Work out partition type, name and flags
+		local $rest = $5;
+		$rest =~ s/^\s+//;
+		$rest =~ s/,//g;	# Remove commas in flags list
+		local @rest = split(/\s+/, $rest);
+
+		# If first word is a known partition type, assume it is the type
+		if (@rest && &conv_type($rest[0])) {
+			$part->{'type'} = shift(@rest);
+			}
+
+		# Remove flag words from the end
+		local %flags;
+		while(@rest && $rest[$#rest] =~ /boot|lba|root|swap|hidden|raid|LVM/i) {
+			$flags{lc(pop(@rest))} = 1;
+			}
+
+		# Anything left in the middle should be the name
+		if (@rest) {
+			$part->{'name'} = $rest[0];
+			}
+		if ($flags{'raid'}) {
+			# RAID flag is set
+			$part->{'raid'} = 1;
+			}
+		$part->{'type'} = 'ext2' if (!$part->{'type'} ||
+					     $part->{'type'} =~ /^ext/);
+		$part->{'type'} = 'raid' if ($part->{'type'} =~ /^ext/ &&
 					     $part->{'raid'});
 		$part->{'desc'} = &partition_description($part->{'device'});
 		$part->{'size'} = ($part->{'end'} - $part->{'start'} + 1) *
@@ -485,7 +590,11 @@ while(<FDISK>) {
 		push(@{$disk->{'parts'}}, $part);
 		}
 	elsif (/Partition\s+Table:\s+(\S+)/) {
-		# Parted partition table type
+		# Parted partition table type (from parted)
+		$disk->{'table'} = $1;
+		}
+	elsif (/Disklabel\s+type:\s+(\S+)/) {
+		# Parted partition table type (from fdisk)
 		$disk->{'table'} = $1;
 		}
 	}
@@ -552,7 +661,7 @@ return @disks;
 }
 
 # partition_description(device)
-# Converts a device path like /dev/hda into a human-readable name
+# Converts a device path like /dev/hda1 into a human-readable name
 sub partition_description
 {
 my ($device) = @_;
@@ -560,6 +669,8 @@ return $device =~ /(s|h|xv|v)d([a-z]+)(\d+)$/ ?
 	 &text('select_part', $1 eq 's' ? 'SCSI' :
 			      $1 eq 'xv' ? 'Xen' :
 			      $1 eq 'v' ? 'VirtIO' : 'IDE', uc($2), "$3") :
+       $device =~ /mmcblk(\d+)p(\d+)$/ ?
+	 &text('select_part', 'SD-Card', "$1", "$2") :
        $device =~ /scsi\/host(\d+)\/bus(\d+)\/target(\d+)\/lun(\d+)\/part(\d+)/ ?
 	 &text('select_spart', "$1", "$2", "$3", "$4", "$5") :
        $device =~ /ide\/host(\d+)\/bus(\d+)\/target(\d+)\/lun(\d+)\/part(\d+)/ ?
@@ -572,6 +683,8 @@ return $device =~ /(s|h|xv|v)d([a-z]+)(\d+)$/ ?
 	 &text('select_smartpart', "$1", "$2", "$3") :
        $device =~ /ataraid\/disc(\d+)\/part(\d+)$/ ?
 	 &text('select_ppart', "$1", "$2") :
+       $device =~ /nvme(\d+)n(\d+)p(\d+)$/ ?
+	 &text('select_nvmepart', "$1", "$2", "$3") :
 	 "???";
 }
 
@@ -604,12 +717,13 @@ else {
 # Changes the type of an existing partition
 sub change_type
 {
-&open_fdisk("$_[0]");
+my ($disk, $part, $type) = @_;
+&open_fdisk($disk);
 &wprint("t\n");
 local $rv = &wait_for($fh, 'Partition.*:', 'Selected partition');
-&wprint("$_[1]\n") if ($rv == 0);
+&wprint("$part\n") if ($rv == 0);
 &wait_for($fh, 'Hex.*:');
-&wprint("$_[2]\n");
+&wprint("$type\n");
 &wait_for($fh, 'Command.*:');
 &wprint("w\n"); sleep(1);
 &close_fdisk();
@@ -623,7 +737,7 @@ sub delete_partition
 my ($disk, $part) = @_;
 if ($has_parted) {
 	# Using parted
-	my $cmd = "parted -s ".$disk." rm ".$part;
+	my $cmd = "parted -s ".quotemeta($disk)." rm ".quotemeta($part);
 	my $out = &backquote_logged("$cmd </dev/null 2>&1");
 	if ($?) {
 		&error("$cmd failed : $out");
@@ -652,19 +766,23 @@ my ($disk, $part, $start, $end, $type) = @_;
 if ($has_parted) {
 	# Using parted
 	my $pe = $part > 4 ? "logical" : "primary";
+	my $qdisk = quotemeta($disk);
+	my $qpart = quotemeta($part);
+	my $qstart = quotemeta($start-1);
+	my $qend = quotemeta($end);
 	my $cmd;
 	if ($type eq "raid") {
-		$cmd = "parted -s ".$disk." unit cyl mkpartfs ".$pe." ".
-		       "ext2 ".($start-1)." ".$end;
-		$cmd .= " ; parted -s ".$disk." set $part raid on";
+		$cmd = "parted -s ".$qdisk." unit cyl mkpart ".$pe." ".
+		       "ext2 ".$qstart." ".$qend;
+		$cmd .= " ; parted -s ".$qdisk." set $qpart raid on";
 		}
 	elsif ($type && $type ne 'ext2') {
-		$cmd = "parted -s ".$disk." unit cyl mkpartfs ".$pe." ".
-		       $type." ".($start-1)." ".$end;
+		$cmd = "parted -s ".$qdisk." unit cyl mkpart ".$pe." ".
+		       quotemeta($type)." ".$qstart." ".$qend;
 		}
 	else {
-		$cmd = "parted -s ".$disk." unit cyl mkpart ".$pe." ".
-		       ($start-1)." ".$end;
+		$cmd = "parted -s ".$qdisk." unit cyl mkpart ".$pe." ".
+		       $qstart." ".$qend;
 		}
 	my $out = &backquote_logged("$cmd </dev/null 2>&1");
 	if ($?) {
@@ -687,16 +805,22 @@ else {
 		}
 	&wait_for($fh, 'First.*:') if ($wf != 1);
 	&wprint("$start\n");
-	&wait_for($fh, 'Last.*:');
+	$wf = &wait_for($fh, 'Last.*:', 'First.*:');
+	$wf < 0 && &error("End of input waiting for first cylinder response");
+	$wf == 1 && &error("First cylinder is invalid : $wait_for_input");
 	&wprint("$end\n");
-	&wait_for($fh, 'Command.*:');
+	$wf = &wait_for($fh, 'Command.*:', 'Last.*:');
+	$wf < 0 && &error("End of input waiting for last cylinder response");
+	$wf == 1 && &error("Last cylinder is invalid : $wait_for_input");
 
 	&wprint("t\n");
 	local $rv = &wait_for($fh, 'Partition.*:', 'Selected partition');
 	&wprint("$part\n") if ($rv == 0);
 	&wait_for($fh, 'Hex.*:');
 	&wprint("$type\n");
-	&wait_for($fh, 'Command.*:');
+	$wf = &wait_for($fh, 'Command.*:', 'Hex.*:');
+	$wf < 0 && &error("End of input waiting for partition type response");
+	$wf == 1 && &error("Partition type is invalid : $wait_for_input");
 	&wprint("w\n");
 	&wait_for($fh, 'Syncing'); sleep(3);
 	&close_fdisk();
@@ -711,8 +835,8 @@ sub create_extended
 my ($disk, $part, $start, $end) = @_;
 if ($has_parted) {
 	# Create using parted
-	my $cmd = "parted -s ".$disk." unit cyl mkpart extended ".
-		  ($start-1)." ".$end;
+	my $cmd = "parted -s ".quotemeta($disk)." unit cyl mkpart ".
+		  "extended ".quotemeta($start-1)." ".quotemeta($end);
 	my $out = &backquote_logged("$cmd </dev/null 2>&1");
 	if ($?) {
 		&error("$cmd failed : $out");
@@ -744,7 +868,12 @@ undef(@list_disks_partitions_cache);
 # Returns a list of known partition tag numbers
 sub list_tags
 {
-if ($has_parted) {
+my ($fdisk_gpt) = @_;
+if ($fdisk_gpt eq 'gpt') {
+	# fdisk gpt types
+	return sort { $a cmp $b } (keys %fdisk_gpt_tags);
+	}
+elsif ($has_parted) {
 	# Parted types
 	return sort { $a cmp $b } (keys %parted_tags);
 	}
@@ -758,7 +887,7 @@ else {
 # Returns a human-readable version of a tag
 sub tag_name
 {
-return $tags{$_[0]} || $parted_tags{$_[0]} || $hidden_tags{$_[0]};
+return $tags{$_[0]} || $parted_tags{$_[0]} || $fdisk_gpt_tags{$_[0]} || $hidden_tags{$_[0]};
 }
 
 sub default_tag
@@ -780,10 +909,10 @@ if ($has_parted) {
 	elsif ($tag eq "fat32") {
 		@rv = ( "vfat" );
 		}
-	elsif ($tag eq "ext2" || $tag eq "raid") {
-		@rv = ( "ext3", "ext4", "ext2", "xfs", "reiserfs" );
+	elsif ($tag =~ /^ext/ || $tag eq "raid") {
+		@rv = ( "ext3", "ext4", "ext2", "xfs", "reiserfs", "btrfs" );
 		}
-	elsif ($tag eq "hfs") {
+	elsif ($tag eq "hfs" || $tag eq "HFS") {
 		@rv = ( "hfs" );
 		}
 	elsif ($tag eq "linux-swap") {
@@ -811,7 +940,7 @@ else {
 		@rv = ( "vfat" );
 		}
 	elsif ($tag eq "83") {
-		@rv = ( "ext3", "ext4", "ext2", "xfs", "reiserfs" );
+		@rv = ( "ext3", "ext4", "ext2", "xfs", "reiserfs", "btrfs" );
 		}
 	elsif ($tag eq "82") {
 		@rv = ( "swap" );
@@ -837,16 +966,7 @@ return $text{"fs_".$_[0]};
 
 sub mkfs_options
 {
-if ($_[0] eq "ext2") {
-	&opt_input("ext2_b", $text{'bytes'}, 1);
-	&opt_input("ext2_f", $text{'bytes'}, 0);
-	&opt_input("ext2_i", "", 1);
-	&opt_input("ext2_m", "%", 0);
-	&opt_input("ext2_g", "", 1);
-	print &ui_table_row($text{'ext2_c'},
-		&ui_yesno_radio("ext2_c", 0));
-	}
-elsif ($_[0] eq "msdos" || $_[0] eq "vfat") {
+if ($_[0] eq "msdos" || $_[0] eq "vfat") {
 	&opt_input("msdos_ff", "", 1);
 	print &ui_table_row($text{'msdos_F'},
 	     &ui_select("msdos_F", undef,
@@ -900,6 +1020,11 @@ elsif ($_[0] eq "fatx") {
 	# Has no options!
 	print &ui_table_row(undef, $text{'fatx_none'}, 4);
 	}
+elsif ($_[0] eq "btrfs") {
+	&opt_input("btrfs_l", $text{'bytes'}, 0);
+	&opt_input("btrfs_n", $text{'bytes'}, 0);
+	&opt_input("btrfs_s", $text{'bytes'}, 0);
+	}
 }
 
 # mkfs_parse(type, device)
@@ -907,36 +1032,26 @@ elsif ($_[0] eq "fatx") {
 # given device. Options are taken from %in.
 sub mkfs_parse
 {
-local($cmd);
-if ($_[0] eq "ext2") {
-	$cmd = "mkfs -t ext2";
-	$cmd .= &opt_check("ext2_b", '\d+', "-b");
-	$cmd .= &opt_check("ext2_f", '\d+', "-f");
-	$cmd .= &opt_check("ext2_i", '\d{4,}', "-i");
-	$cmd .= &opt_check("ext2_m", '\d+', "-m");
-	$cmd .= &opt_check("ext2_g", '\d+', "-g");
-	$cmd .= $in{'ext2_c'} ? " -c" : "";
-	$cmd .= " -q";
-	$cmd .= " $_[1]";
-	}
-elsif ($_[0] eq "msdos" || $_[0] eq "vfat") {
+local($cmd, $qdev);
+$qdev = quotemeta($_[1]);
+if ($_[0] eq "msdos" || $_[0] eq "vfat") {
 	$cmd = "mkfs -t $_[0]";
 	$cmd .= &opt_check("msdos_ff", '[1-2]', "-f");
 	if ($in{'msdos_F'} eq '*') {
 		$in{'msdos_F_other'} =~ /^\d+$/ ||
 			&error(&text('opt_error', $in{'msdos_F_other'},
 						  $text{'msdos_F'}));
-		$cmd .= " -F ".$in{'msdos_F_other'};
+		$cmd .= " -F ".quotemeta($in{'msdos_F_other'});
 		}
 	elsif ($in{'msdos_F'}) {
-		$cmd .= " -F ".$in{'msdos_F'};
+		$cmd .= " -F ".quotemeta($in{'msdos_F'});
 		}
 	$cmd .= &opt_check("msdos_i", '[0-9a-f]{8}', "-i");
 	$cmd .= &opt_check("msdos_n", '\S{1,11}', "-n");
 	$cmd .= &opt_check("msdos_r", '\d+', "-r");
 	$cmd .= &opt_check("msdos_s", '\d+', "-s");
 	$cmd .= $in{'msdos_c'} ? " -c" : "";
-	$cmd .= " $_[1]";
+	$cmd .= " $qdev";
 	}
 elsif ($_[0] eq "minix") {
 	local(@plist, $disk, $part, $i, @pinfo);
@@ -945,13 +1060,13 @@ elsif ($_[0] eq "minix") {
 	$cmd .= &opt_check("minix_i", '\d+', "-i ");
 	$cmd .= $in{'minix_c'} ? " -c" : "";
 	$cmd .= &opt_check("minix_b", '\d+', " ");
-	$cmd .= " $_[1]";
+	$cmd .= " $qdev";
 	}
 elsif ($_[0] eq "reiserfs") {
 	$cmd = "yes | mkreiserfs";
 	$cmd .= " -f" if ($in{'reiserfs_f'});
-	$cmd .= " -h $in{'reiserfs_h'}" if ($in{'reiserfs_h'});
-	$cmd .= " $_[1]";
+	$cmd .= " -h ".quotemeta($in{'reiserfs_h'}) if ($in{'reiserfs_h'});
+	$cmd .= " $qdev";
 	}
 elsif ($_[0] =~ /^ext\d+$/) {
 	if (&has_command("mkfs.$_[0]")) {
@@ -972,7 +1087,7 @@ elsif ($_[0] =~ /^ext\d+$/) {
 			$in{'ext3_j'} =~ /^\d+$/ ||
 				&error(&text('opt_error', $in{'ext3_j'},
 					     $text{'ext3_j'}));
-			$cmd .= " -J size=$in{'ext3_j'}";
+			$cmd .= " -J size=".quotemeta($in{'ext3_j'});
 			}
 		}
 	$cmd .= &opt_check("ext2_b", '\d+', "-b");
@@ -981,22 +1096,29 @@ elsif ($_[0] =~ /^ext\d+$/) {
 	$cmd .= &opt_check("ext2_m", '\d+', "-m");
 	$cmd .= $in{'ext2_c'} ? " -c" : "";
 	$cmd .= " -q";
-	$cmd .= " $_[1]";
+	$cmd .= " $qdev";
 	}
 elsif ($_[0] eq "xfs") {
 	$cmd = "mkfs -t $_[0]";
 	$cmd .= " -f" if ($in{'xfs_f'});
-	$cmd .= " -b size=$in{'xfs_b'}" if (!$in{'xfs_b_def'});
-	$cmd .= " $_[1]";
+	$cmd .= " -b size=".quotemeta($in{'xfs_b'}) if (!$in{'xfs_b_def'});
+	$cmd .= " $qdev";
 	}
 elsif ($_[0] eq "jfs") {
 	$cmd = "mkfs -t $_[0] -q";
 	$cmd .= &opt_check("jfs_s", '\d+', "-s");
 	$cmd .= " -c" if ($in{'jfs_c'});
-	$cmd .= " $_[1]";
+	$cmd .= " $qdev";
 	}
 elsif ($_[0] eq "fatx") {
-	$cmd = "mkfs -t $_[0] $_[1]";
+	$cmd = "mkfs -t $_[0] $qdev";
+	}
+elsif ($_[0] eq "btrfs") {
+	$cmd = "mkfs -t $_[0]";
+	$cmd .= " -l ".quotemeta($in{'btrfs_l'}) if (!$in{'btrfs_l_def'});
+	$cmd .= " -n ".quotemeta($in{'btrfs_n'}) if (!$in{'btrfs_n_def'});
+	$cmd .= " -s ".quotemeta($in{'btrfs_s'}) if (!$in{'btrfs_s_def'});
+	$cmd .= " $qdev";
 	}
 if (&has_command("partprobe")) {
 	$cmd = "partprobe ; $cmd";
@@ -1054,16 +1176,17 @@ if ($_[0] =~ /^ext\d+$/) {
 # Returns the tuning command based on user inputs
 sub tunefs_parse
 {
+local $qdev = quotemeta($_[1]);
 if ($_[0] =~ /^ext\d+$/) {
 	$cmd = "tune2fs";
 	$cmd .= &opt_check("tunefs_c", '\d+', "-c");
-	$cmd .= $in{'tunefs_e_def'} ? "" : " -e$in{'tunefs_e'}";
-	$cmd .= $in{'tunefs_u_def'} ? "" : " -u".getpwnam($in{'tunefs_u'});
-	$cmd .= $in{'tunefs_g_def'} ? "" : " -g".getgrnam($in{'tunefs_g'});
+	$cmd .= $in{'tunefs_e_def'} ? "" : " -e".quotemeta($in{'tunefs_e'});
+	$cmd .= $in{'tunefs_u_def'} ? "" : " -u".quotemeta(getpwnam($in{'tunefs_u'}));
+	$cmd .= $in{'tunefs_g_def'} ? "" : " -g".quotemeta(getgrnam($in{'tunefs_g'}));
 	$cmd .= &opt_check("tunefs_m",'\d+',"-m");
 	$cmd .= &opt_check("tunefs_i", '\d+', "-i").
-		($in{'tunefs_i_def'} ? "" : $in{'tunefs_i_unit'});
-	$cmd .= " $_[1]";
+		($in{'tunefs_i_def'} ? "" : quotemeta($in{'tunefs_i_unit'}));
+	$cmd .= " $qdev";
 	}
 return $cmd;
 }
@@ -1077,7 +1200,7 @@ return $un =~ /^2\.0\./ || $un =~ /^1\./ || $un =~ /^0\./;
 }
 
 # device_status(device)
-# Returns an array of  directory, type, mounted
+# Returns an array of  directory, type, mounted, module
 sub device_status
 {
 @mounted = &foreign_call("mount", "list_mounted") if (!@mounted);
@@ -1100,7 +1223,8 @@ if ($raid_module) {
 	my $raidconf = &foreign_call("raid", "get_raidtab") if (!$raidconf);
 	foreach $c (@$raidconf) {
 		foreach $d (&raid::find_value('device', $c->{'members'})) {
-			return ( $c->{'value'}, "raid", 1 ) if ($d eq $_[0]);
+			return ( $c->{'value'}, "raid", 1, "raid" )
+				if ($d eq $_[0]);
 			}
 		}
 	}
@@ -1114,21 +1238,79 @@ if ($lvm_module) {
 			}
 		}
 	foreach my $pv (@physical_volumes) {
-		return ( $pv->{'vg'}, "lvm", 1)
+		return ( $pv->{'vg'}, "lvm", 1, "lvm")
 			if ($pv->{'device'} eq $_[0]);
 		}
 	}
-if ($iscsi_module) {
+if ($iscsi_server_module) {
 	my $iscsiconf = &iscsi_server::get_iscsi_config();
 	foreach my $c (@$iscsiconf) {
 		if ($c->{'type'} eq 'extent' && $c->{'device'} eq $_[0]) {
-			return ( $c->{'type'}.$c->{'num'}, "iscsi", 1);
+			return ( $c->{'type'}.$c->{'num'}, "iscsi", 1,
+				 "iscsi-server");
+			}
+		}
+	}
+if ($iscsi_target_module) {
+	my $iscsiconf = &iscsi_target::get_iscsi_config();
+	foreach my $t (&iscsi_target::find($iscsiconf, "Target")) {
+		foreach my $l (&iscsi_target::find($t->{'members'}, "Lun")) {
+			if ($l->{'value'} =~ /Path=([^, ]+)/ && $1 eq $_[0]) {
+				return ( $t->{'value'}, "iscsi", 1,
+					 "iscsi-target");
+				}
 			}
 		}
 	}
 return ();
 }
 
+# device_status_link(directory, type, mounted, module)
+# Converts the list returned by device_status to a link
+sub device_status_link
+{
+my @stat = @_;
+my $stat = "";
+my $statdesc = $stat[0] =~ /^swap/ ? "<i>$text{'disk_vm'}</i>"
+				   : "<tt>$stat[0]</tt>";
+my $ret = $main::initial_module_name;
+if ($ret !~ /fdisk$/) {
+	$ret = $module_name;
+	}
+if ($stat[1] eq 'raid') {
+	$stat = $statdesc;
+	}
+elsif ($stat[1] eq 'lvm') {
+	if (&foreign_available("lvm")) {
+		$stat = "<a href='../lvm/'>".
+			"LVM VG $statdesc</a>";
+		}
+	else {
+		$stat = "LVM VG $statdesc";
+		}
+	}
+elsif ($stat[1] eq 'iscsi') {
+	$stat = &text('disk_iscsi', $stat[0]);
+	if (&foreign_available("iscsi-server")) {
+		$stat = "<a href='../$stat[3]/'>$stat</a>";
+		}
+	}
+elsif ($stat[0] && !&foreign_available("mount")) {
+	$stat = $statdesc;
+	}
+elsif ($stat[0] && $stat[3] == -1) {
+	$stat = "<a href='../mount/edit_mount.cgi?".
+		"index=$stat[4]&temp=1&return=/$ret/'>".
+		"$statdesc</a>";
+	}
+elsif ($stat[0]) {
+	$stat = "<a href='../mount/edit_mount.cgi?".
+		"index=$stat[3]&return=/$ret/'>".
+		"$statdesc</a>";
+	}
+return $stat;
+}
+
 # can_fsck(type)
 # Returns 1 if some filesystem type can fsck'd
 sub can_fsck
@@ -1141,11 +1323,12 @@ return ($_[0] =~ /^ext\d+$/ && &has_command("fsck.$_[0]") ||
 # Returns the fsck command to unconditionally check a filesystem
 sub fsck_command
 {
+local $qdev = quotemeta($_[1]);
 if ($_[0] =~ /^ext\d+$/) {
-	return "fsck -t $_[0] -p $_[1]";
+	return "fsck -t $_[0] -p $qdev";
 	}
 elsif ($_[0] eq "minix") {
-	return "fsck -t minix -a $_[1]";
+	return "fsck -t minix -a $qdev";
 	}
 }
 
@@ -1240,7 +1423,7 @@ foreach my $d (@dlist) {
 	foreach $p (@{$d->{'parts'}}) {
 		next if ($p->{'type'} ne '83' && $p->{'type'} ne '82' &&
 			 $p->{'type'} ne 'b' && $p->{'type'} ne 'c' &&
-			 $p->{'type'} !~ /^ext/);
+			 $p->{'type'} !~ /^(ext|xfs)/);
 		local $volid = &get_volid($p->{'device'});
 		next if (!$volid);
 		push(@opts, [ $volid, "$volid ($p->{'desc'})" ]);
@@ -1261,13 +1444,27 @@ else {
 sub open_fdisk
 {
 local $fpath = &check_fdisk();
-($fh, $fpid) = &foreign_call("proc", "pty_process_exec", join(" ",$fpath, @_));
+local $qfpath = quotemeta($fpath);
+my $cylarg;
+if ($fpath =~ /\/fdisk/) {
+	my $out = &backquote_command("$qfpath -h 2>&1 </dev/null");
+	if ($out =~ /-u\s+<size>/) {
+		$cylarg = "-u=cylinders";
+		}
+	}
+local @qargs = map { quotemeta($_) } grep { defined($_) && $_ ne "" }
+	       ($fpath, $cylarg, @_);
+($fh, $fpid) = &foreign_call("proc", "pty_process_exec",
+			     join(" ", @qargs));
 }
 
 sub open_sfdisk
 {
 local $sfpath = &has_command("sfdisk");
-($fh, $fpid) = &foreign_call("proc", "pty_process_exec", join(" ",$sfpath, @_));
+local @qargs = map { quotemeta($_) } grep { defined($_) && $_ ne "" }
+	       ($sfpath, @_);
+($fh, $fpid) = &foreign_call("proc", "pty_process_exec",
+			     join(" ", @qargs));
 }
 
 sub check_fdisk
@@ -1299,7 +1496,7 @@ if ($in{"$_[0]_def"}) { return ""; }
 elsif ($in{$_[0]} !~ /^$_[1]$/) {
 	&error(&text('opt_error', $in{$_[0]}, $text{$_[0]}));
 	}
-else { return " $_[2] $in{$_[0]}"; }
+else { return " $_[2] ".quotemeta($in{$_[0]}); }
 }
 
 %tags = ('0', 'Empty',
@@ -1387,11 +1584,20 @@ else { return " $_[2] $in{$_[0]}"; }
 	 'f', 'Windows extended LBA',
 	);
 
+%fdisk_gpt_tags = (
+	'', 'None',
+	'EFI System', 'EFI system partition',
+	'BIOS boot', 'BIOS boot partition',
+	'Linux filesystem', 'Linux filesystem',
+	'Linux LVM', 'Linux LVM',
+	);
+
 %parted_tags = (
 	'', 'None',
 	'fat16', 'Windows FAT16',
 	'fat32', 'Windows FAT32',
-	'ext2', 'Linux',
+	'ext2', 'Linux EXT',
+	'xfs', 'Linux XFS',
 	'raid', 'Linux RAID',
 	'HFS', 'MacOS HFS',
 	'linux-swap', 'Linux Swap',
@@ -1419,7 +1625,8 @@ return 0;
 sub disk_space
 {
 local $w = $_[1] || $_[0];
-local $out = `df -k '$w'`;
+local $qw = quotemeta($w);
+local $out = `df -k $qw`;
 if ($out =~ /Mounted on\s*\n\s*\S+\s+(\S+)\s+\S+\s+(\S+)/i) {
 	return ($1, $2);
 	}
@@ -1445,6 +1652,7 @@ push(@fstypes, "reiserfs") if (&has_command("mkreiserfs"));
 push(@fstypes, "xfs") if (&has_command("mkfs.xfs"));
 push(@fstypes, "jfs") if (&has_command("mkfs.jfs"));
 push(@fstypes, "fatx") if (&has_command("mkfs.fatx"));
+push(@fstypes, "btrfs") if (&has_command("mkfs.btrfs"));
 push(@fstypes, "msdos");
 push(@fstypes, "vfat");
 push(@fstypes, "minix");
@@ -1456,19 +1664,21 @@ return @fstypes;
 sub get_label
 {
 local $label;
+local $qdev = quotemeta($_[0]);
 if ($has_e2label) {
-	$label = `e2label $_[0] 2>&1`;
+	$label = `e2label $qdev 2>&1`;
 	chop($label);
 	}
 if (($? || $label !~ /\S/) && $has_xfs_db) {
 	$label = undef;
-	local $out = &backquote_with_timeout("xfs_db -x -p xfs_admin -c label -r $_[0] 2>&1", 5);
+	local $out = &backquote_with_timeout(
+		"xfs_db -x -p xfs_admin -c label -r $qdev 2>&1", 5);
 	$label = $1 if ($out =~ /label\s*=\s*"(.*)"/ &&
 			$1 ne '(null)');
 	}
 if (($? || $label !~ /\S/) && $has_reiserfstune) {
 	$label = undef;
-	local $out = &backquote_command("reiserfstune $_[0]");
+	local $out = &backquote_command("reiserfstune $qdev");
 	if ($out =~ /LABEL:\s*(\S+)/) {
 		$label = $1;
 		}
@@ -1510,12 +1720,15 @@ return $uuid;
 # Tries to set the label for some device's filesystem
 sub set_label
 {
+local $qdev = quotemeta($_[0]);
+local $qlabel = quotemeta($_[1]);
 if ($has_e2label && ($_[2] =~ /^ext[23]$/ || !$_[2])) {
-	&system_logged("e2label '$_[0]' '$_[1]' >/dev/null 2>&1");
+	&system_logged("e2label $qdev $qlabel >/dev/null 2>&1");
 	return 1 if (!$?);
 	}
 if ($has_xfs_db && ($_[2] eq "xfs" || !$_[2])) {
-	&system_logged("xfs_db -x -p xfs_admin -c \"label $_[1]\" $_[0] >/dev/null 2>&1");
+	&system_logged(
+		"xfs_db -x -p xfs_admin -c \"label $qlabel\" $qdev >/dev/null 2>&1");
 	return 1 if (!$?);
 	}
 return 0;
@@ -1526,7 +1739,8 @@ return 0;
 sub set_name
 {
 my ($dinfo, $pinfo, $name) = @_;
-my $cmd = "parted -s ".$dinfo->{'device'}." name ".$pinfo->{'number'}." ";
+my $cmd = "parted -s ".quotemeta($dinfo->{'device'})." name ".
+	  quotemeta($pinfo->{'number'})." ";
 if ($name) {
 	$cmd .= quotemeta($name);
 	}
@@ -1544,7 +1758,7 @@ if ($?) {
 sub set_partition_table
 {
 my ($disk, $table) = @_;
-my $cmd = "parted -s ".$disk." mktable ".$table;
+my $cmd = "parted -s ".quotemeta($disk)." mktable ".quotemeta($table);
 my $out = &backquote_logged("$cmd </dev/null 2>&1");
 if ($?) {
 	&error("$cmd failed : $out");
@@ -1571,7 +1785,8 @@ return $disk->{'table'} eq 'gpt';
 sub supports_hdparm
 {
 local ($d) = @_;
-return $d->{'type'} eq 'ide' || $d->{'type'} eq 'scsi' && $d->{'model'} =~ /ATA/;
+return $d->{'type'} eq 'ide' ||
+       $d->{'type'} eq 'scsi' && $d->{'model'} =~ /ATA/;
 }
 
 # supports_relabel(&disk)
